﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Microsoft.Kinect;

namespace KinectSimpleSample
{
    public partial class KinectSimpleSample : Form
    {
        // Kinectクラス
        private KinectSensor Kinect;

        // 表示用フォーム
        private frmColorViewer ColorViewerForm = new frmColorViewer();
        private frmDepthViewer DepthViewerForm = new frmDepthViewer();

        // データフォーマット保持用
        private ColorImageFormat lastImageFormat = ColorImageFormat.Undefined;
        private DepthImageFormat DepthLastImageFormat = DepthImageFormat.Undefined;
        
        // カラーデータ格納用配列
        private byte[] pixelData;
        // Depthデータ格納用配列
        private short[] DepthPixelData;
        private byte[] depthFrame32;
        // Skeletonデータ格納用配列
        private Skeleton[] skeletonData;
       
        // color divisors for tinting depth pixels
        private static readonly int[] IntensityShiftByPlayerR = { 1, 2, 0, 2, 0, 0, 2, 0 };
        private static readonly int[] IntensityShiftByPlayerG = { 1, 2, 2, 0, 2, 0, 0, 1 };
        private static readonly int[] IntensityShiftByPlayerB = { 1, 0, 2, 2, 0, 2, 0, 2 };

        private const int RedIndex = 2;
        private const int GreenIndex = 1;
        private const int BlueIndex = 0;
        private static readonly int Bgr32BytesPerPixel = 4;//(PixelFormats.Bgr32.BitsPerPixel + 7) / 8;

        public KinectSimpleSample()
        {
            InitializeComponent();
        }

        private void KinectSimpleSample_Load(object sender, EventArgs e)
        {
            // Kinectクラス
            Kinect = (from sensorToCheck in KinectSensor.KinectSensors
                                   where sensorToCheck.Status == KinectStatus.Connected
                                   select sensorToCheck).FirstOrDefault();

            if (Kinect != null)
            {
                // 表示フォームの表示
                ColorViewerForm.Show();
                DepthViewerForm.Show();

                // カラー画像を有効にする。（フォーマットの設定）
                Kinect.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
                // Depthデータを有効にする。（フォーマットの設定）
                Kinect.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);
                // スケルトンを有効にする。
                Kinect.SkeletonStream.Enable();

                // ElevationAngle(首ふり)の設定を取得
                trbElevationAngle.Minimum = Kinect.MinElevationAngle;
                trbElevationAngle.Maximum = Kinect.MaxElevationAngle;

                // Kinectの開始
                Kinect.Start();

                // イベントの設定
                Kinect.ColorFrameReady += this.ColorImageReady;
                Kinect.DepthFrameReady += this.DepthImageReady;
                Kinect.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KinectAllFramesReady);
            }
        }

        private void KinectSimpleSample_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (Kinect != null)
            {
                // Kinectの停止
                Kinect.Stop();
                Kinect = null;
            }
        }

        private void trbElevationAngle_MouseUp(object sender, MouseEventArgs e)
        {
            // ElevationAngle(首ふり)の設定
            Kinect.ElevationAngle = trbElevationAngle.Value;
        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            // 終了
            this.Close();
        }

        // カラー画像を取得した時のイベント
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame imageFrame = e.OpenColorImageFrame())
            {
                if (imageFrame != null)
                {
                    // 取込フォーマット（サイズ）が変更されたか？
                    bool haveNewFormat = this.lastImageFormat != imageFrame.Format;

                    // フォーマットが変更された場合、メモリを確保
                    if (haveNewFormat)
                    {
                        // カラー画像格納用配列の確保
                        this.pixelData = new byte[imageFrame.PixelDataLength];
                    }

                    // カラーデータを配列へコピー
                    imageFrame.CopyPixelDataTo(this.pixelData);

                    unsafe
                    {
                        fixed (byte* pData = this.pixelData)
                        {
                            Bitmap bmp = new Bitmap(imageFrame.Width, imageFrame.Height,
                                imageFrame.PixelDataLength / imageFrame.Height,
                                System.Drawing.Imaging.PixelFormat.Format32bppRgb,
                                (IntPtr)pData);
                            ColorViewerForm.picColor.Image = bmp;
                        }
                    }

                    this.lastImageFormat = imageFrame.Format;
                }
            }
        }

        // Depthデータを取得した時のイベント
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame imageFrame = e.OpenDepthImageFrame())
            {
                if (imageFrame != null)
                {
                    // 取込フォーマット（サイズ）が変更されたか？
                    bool haveNewFormat = this.DepthLastImageFormat != imageFrame.Format;

                    // フォーマットが変更された場合、メモリを確保
                    if (haveNewFormat)
                    {
                        // Depthデータ格納用配列の確保
                        this.DepthPixelData = new short[imageFrame.PixelDataLength];
                        this.depthFrame32 = new byte[imageFrame.Width * imageFrame.Height * Bgr32BytesPerPixel];
                    }

                    // Depthデータを配列へコピー
                    imageFrame.CopyPixelDataTo(this.DepthPixelData);

                    byte[] convertedDepthBits = this.ConvertDepthFrame(this.DepthPixelData, ((KinectSensor)sender).DepthStream);

                    unsafe
                    {
                        fixed (byte* pData = this.depthFrame32)
                        {
                            Bitmap bmp = new Bitmap(imageFrame.Width, imageFrame.Height,
                                imageFrame.Width * Bgr32BytesPerPixel,
                                System.Drawing.Imaging.PixelFormat.Format32bppRgb,
                                (IntPtr)pData);
                            DepthViewerForm.picDepth.Image = bmp;
                        }
                    }                  

                    this.DepthLastImageFormat = imageFrame.Format;
                }
            }
        }

        // Depthデータを表示用のデータへ変換
        // Converts a 16-bit grayscale depth frame which includes player indexes into a 32-bit frame
        // that displays different players in different colors
        private byte[] ConvertDepthFrame(short[] depthFrame, DepthImageStream depthStream)
        {
            int tooNearDepth = depthStream.TooNearDepth;
            int tooFarDepth = depthStream.TooFarDepth;
            int unknownDepth = depthStream.UnknownDepth;

            for (int i16 = 0, i32 = 0; i16 < depthFrame.Length && i32 < this.depthFrame32.Length; i16++, i32 += 4)
            {
                int player = depthFrame[i16] & DepthImageFrame.PlayerIndexBitmask;
                int realDepth = depthFrame[i16] >> DepthImageFrame.PlayerIndexBitmaskWidth;

                // transform 13-bit depth information into an 8-bit intensity appropriate
                // for display (we disregard information in most significant bit)
                byte intensity = (byte)(~(realDepth >> 4));

                if (player == 0 && realDepth == 0)
                {
                    // white 
                    this.depthFrame32[i32 + RedIndex] = 255;
                    this.depthFrame32[i32 + GreenIndex] = 255;
                    this.depthFrame32[i32 + BlueIndex] = 255;
                }
                else if (player == 0 && realDepth == tooFarDepth)
                {
                    // dark purple
                    this.depthFrame32[i32 + RedIndex] = 66;
                    this.depthFrame32[i32 + GreenIndex] = 0;
                    this.depthFrame32[i32 + BlueIndex] = 66;
                }
                else if (player == 0 && realDepth == unknownDepth)
                {
                    // dark brown
                    this.depthFrame32[i32 + RedIndex] = 66;
                    this.depthFrame32[i32 + GreenIndex] = 66;
                    this.depthFrame32[i32 + BlueIndex] = 33;
                }
                else
                {
                    // tint the intensity by dividing by per-player values
                    this.depthFrame32[i32 + RedIndex] = (byte)(intensity >> IntensityShiftByPlayerR[player]);
                    this.depthFrame32[i32 + GreenIndex] = (byte)(intensity >> IntensityShiftByPlayerG[player]);
                    this.depthFrame32[i32 + BlueIndex] = (byte)(intensity >> IntensityShiftByPlayerB[player]);
                }
            }

            return this.depthFrame32;
        }

        // Skeletonデータを取得した時のイベント
        private void KinectAllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            // Have we already been "shut down" by the user of this viewer, 
            // or has the SkeletonStream been disabled since this event was posted?
            if ((this.Kinect == null) || !((KinectSensor)sender).SkeletonStream.IsEnabled)
            {
                return;
            }

            bool haveSkeletonData = false;

            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null)
                {
                    if ((this.skeletonData == null) || (this.skeletonData.Length != skeletonFrame.SkeletonArrayLength))
                    {
                        // Skeletonデータ格納用配列の確保
                        this.skeletonData = new Skeleton[skeletonFrame.SkeletonArrayLength];
                    }
                    skeletonFrame.CopySkeletonDataTo(this.skeletonData);
                    haveSkeletonData = true;
                }
            }

            if (haveSkeletonData)
            {
                foreach (Skeleton skeleton in this.skeletonData)
                {
                    Graphics g = Graphics.FromImage(DepthViewerForm.picDepth.Image);

                    foreach (Joint joint in skeleton.Joints)
                    {
                        DepthImagePoint depthPoint = Kinect.MapSkeletonPointToDepth(joint.Position, this.DepthLastImageFormat);


                        if ((depthPoint.X != 0) && (depthPoint.Y != 0))
                        {
                            // Skeleton座標の描画
                            g.FillEllipse(Brushes.Red,
                                depthPoint.X - 4,
                                depthPoint.Y - 4,
                                8, 8);
                        }
                    }
                    DepthViewerForm.picDepth.Refresh();
                }

            }
        }



    }
}
